#pragma once

#include <stdint.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include "../utils/delay.hpp"


namespace Devices
{
    template <class ConnectionT, bool twoLines = false, bool largeFont = false>
    class Hd44780
    {
        public:

            static uint8_t const CMD_CLEAR = 0x01;
            static uint8_t const CMD_HOME = 0x02;

            static uint8_t const CMD_ENTRY_MODE_SET = 0x04;
            static uint8_t const ENTRY_MODE_INCREMENT = 1;
            static uint8_t const ENTRY_MODE_SHIFT = 0;

            static uint8_t const CMD_DISPLAY_CONTROL = 0x08;
            static uint8_t const DISPLAY_CONTROL_DISPLAY_ON = 2;
            static uint8_t const DISPLAY_CONTROL_CURSOR_ON = 1;
            static uint8_t const DISPLAY_CONTROL_BLINKING_ON = 0;

            static uint8_t const CMD_CURSOR_SHIFT = 0x10;

            static uint8_t const CMD_FUNCTION_SET = 0x20;
            static uint8_t const FUNCTION_SET_DATA_LENGTH = 4;
            static uint8_t const FUNCTION_SET_DISPLAY_LINES = 3;
            static uint8_t const FUNCTION_SET_CHARACTER_FONT = 2;

            static uint8_t const CMD_SET_CGRAM_ADDR = 0x40;
            static uint8_t const CMD_SET_DDRAM_ADDR = 0x80;


            static void Init()
            {
                ConnectionT::Init();
                Configure();
            }


            static void Clear()
            {
                SlowCommand(CMD_CLEAR);
            }


            static void Home()
            {
                SlowCommand(CMD_HOME);
            }


            static void EntryModeSet(bool const increment, bool const shift)
            {
                uint8_t cmd = CMD_ENTRY_MODE_SET;

                if (increment)
                {
                    cmd |= (1 << ENTRY_MODE_INCREMENT);
                }
                
                if (shift)
                {
                    cmd |= (1 << ENTRY_MODE_SHIFT);
                }
                
                Command(cmd);
            }


            static void FunctionSet(bool const displayLines, bool const characterFont)
            {
                uint8_t cmd = CMD_FUNCTION_SET;

                if (!ConnectionT::halfWidth)
                {
                    cmd |= (1 << FUNCTION_SET_DATA_LENGTH);
                }
                
                if (displayLines)
                {
                    cmd |= (1 << FUNCTION_SET_DISPLAY_LINES);
                }
                
                if (characterFont)
                {
                    cmd |= (1 << FUNCTION_SET_CHARACTER_FONT);
                }

                Command(cmd);
            }


            static void DisplayControl(bool const displayOn, bool const cursorOn, bool const blinkingOn)
            {
                uint8_t cmd = CMD_DISPLAY_CONTROL;

                if (displayOn)
                {
                    cmd |= (1 << DISPLAY_CONTROL_DISPLAY_ON);
                }
                
                if (cursorOn)
                {
                    cmd |= (1 << DISPLAY_CONTROL_CURSOR_ON);
                }

                if (blinkingOn)
                {
                    cmd |= (1 << DISPLAY_CONTROL_BLINKING_ON);
                }

                Command(cmd);
            }


            static void Move(uint8_t const x, uint8_t const y) 
            {
                Command(CMD_SET_DDRAM_ADDR | (x + y * 0x40));
            }


            static void PutChar(char const ch) 
            {
                Data(ch);
            }


            static void LoadCharProgmem(uint8_t const charIdx, uint8_t const * const charData) 
            {
                Command(CMD_SET_CGRAM_ADDR | (charIdx << 3));

                for (uint8_t i = 0; i < 8; ++i)
                {
                    Data(pgm_read_byte(charData + i));
                }

                Move(0, 0);
            }

            static void SetBacklight(bool const isOn)
            {
                ConnectionT::SetBacklight(isOn);
            }

        private:

            constexpr static uint16_t const FAST_CMD_DELAY_US = 50;
            constexpr static uint16_t const SLOW_CMD_DELAY_US = 2000;


            static void Write(uint8_t data)
            {
                if (ConnectionT::halfWidth)
                {
                    ConnectionT::Out(data >> 4);
                    ConnectionT::SetEn(true);
                    ConnectionT::SetEn(false);
                    ConnectionT::Out(data);
                    ConnectionT::SetEn(true);
                    ConnectionT::SetEn(false);
                }
                else
                {
                    ConnectionT::Out(data);
                    ConnectionT::SetEn(true);
                    ConnectionT::SetEn(false);
                }
            }


            static uint8_t Read()
            {
                uint8_t data;

                if (ConnectionT::halfWidth)
                {
                    ConnectionT::SetEn(true);
                    data = ConnectionT::In() << 4;
                    ConnectionT::SetEn(false);

                    ConnectionT::SetEn(true);
                    data |= ConnectionT::In();
                    ConnectionT::SetEn(false);
                }
                else
                {
                    ConnectionT::SetEn(true);
                    data = ConnectionT::In();
                    ConnectionT::SetEn(false);
                }

                return data;
            }


            static void Wait()
            {
                ConnectionT::SetRw(true);
                ConnectionT::SetRs(false);

                while (Read() & 0x80);

                ConnectionT::SetRw(false);
            }


            static void Command(uint8_t code)
            {
                if (ConnectionT::canRead)
                {
                    Wait();
                }

                ConnectionT::SetRs(false);
                Write(code);

                if (!ConnectionT::canRead)
                {
                    Utils::DelayUs(FAST_CMD_DELAY_US);
                }
            }


            static void SlowCommand(uint8_t code)
            {
                Command(code);

                if (!ConnectionT::canRead)
                {
                    Utils::DelayUs(SLOW_CMD_DELAY_US);
                }
            }


            static void Data(uint8_t data)
            {
                if (ConnectionT::canRead)
                {
                    Wait();
                }

                ConnectionT::SetRs(true);
                Write(data);
                ConnectionT::SetRs(false);

                if (!ConnectionT::canRead)
                {
                    Utils::DelayUs(FAST_CMD_DELAY_US);
                }
            }

            static void Configure()
            {
                ConnectionT::SetEn(false);
                ConnectionT::SetRs(false);
                ConnectionT::SetRw(false);

                uint8_t initCmd = CMD_FUNCTION_SET | (1 << FUNCTION_SET_DATA_LENGTH);

                if (ConnectionT::halfWidth)
                {
                    initCmd >>= 4;
                }

                Utils::DelayMs(15);
                
                ConnectionT::Out(initCmd);
                ConnectionT::SetEn(true);
                ConnectionT::SetEn(false);
                Utils::DelayUs(4100);

                ConnectionT::SetEn(true);
                ConnectionT::SetEn(false);
                Utils::DelayUs(100);

                ConnectionT::SetEn(true);
                ConnectionT::SetEn(false);
                Utils::DelayUs(100);

                if (ConnectionT::halfWidth)
                {
                    ConnectionT::Out(CMD_FUNCTION_SET >> 4);
                    ConnectionT::SetEn(true);
                    ConnectionT::SetEn(false);
                    Utils::DelayUs(FAST_CMD_DELAY_US);
                }

                if (twoLines || largeFont)
                {
                    FunctionSet(twoLines, largeFont);
                }

                DisplayControl(true, false, false);
                EntryModeSet(true, false);
                Clear();
            }
    };

    template <class RsPin, class EnPin, class D4Pin, class D5Pin, class D6Pin, class D7Pin>
    class Hd44780Transport6Pin
    {
        public:

            constexpr static bool const canRead = false;
            constexpr static bool const halfWidth = true;

            static void Init()
            {
                RsPin::ConfigureAsOutput();
                EnPin::ConfigureAsOutput();
                D7Pin::ConfigureAsOutput();
                D6Pin::ConfigureAsOutput();
                D5Pin::ConfigureAsOutput();
                D4Pin::ConfigureAsOutput();
            }

            static void SetEn(bool const value)
            {
                EnPin::Set(value);
            }

            static void SetRs(bool const value)
            {
                RsPin::Set(value);
            }

            static void SetRw(bool const value)
            {
            }

            static void Out(uint8_t const data)
            {
                D7Pin::Set(data & 0x08);
                D6Pin::Set(data & 0x04);
                D5Pin::Set(data & 0x02);
                D4Pin::Set(data & 0x01);
            }

            static uint8_t In()
            {
                return 0;
            }
    };

    template <class RegisterT, uint8_t RsPin, uint8_t RwPin, uint8_t EnPin, uint8_t D4Pin, uint8_t D5Pin, uint8_t D6Pin, uint8_t D7Pin, uint8_t BacklightPin>
    class Hd44780TransportRegister
    {
        private:

            static uint8_t registerValue;

        public:

            using Register = RegisterT;

            constexpr static bool const canRead = false;
            constexpr static bool const halfWidth = true;

            static void Init()
            {
                registerValue = 0;
                RegisterT::Init();
                RegisterT::Out(registerValue);
            }

            static void SetEn(bool const value)
            {
                if (value)
                {
                    RegisterT::Out(registerValue);
                    registerValue |= (1 << EnPin);
                }
                else
                {
                    registerValue &= ~(1 << EnPin);
                }

                RegisterT::Out(registerValue);
            }

            static void SetRs(bool const value)
            {
                if (value)
                {
                    registerValue |= (1 << RsPin);
                }
                else
                {
                    registerValue &= ~(1 << RsPin);
                }
                
                // RegisterT::Out(registerValue);
            }

            static void SetRw(bool const value)
            {
                if (value)
                {
                    registerValue |= (1 << RwPin);
                }
                else
                {
                    registerValue &= ~(1 << RwPin);
                }
            }

            static void Out(uint8_t const byte)
            {
                registerValue &= ~(1 << D7Pin);
                registerValue &= ~(1 << D6Pin);
                registerValue &= ~(1 << D5Pin);
                registerValue &= ~(1 << D4Pin);

                if (byte & 0x08)
                {
                    registerValue |= (1 << D7Pin);
                }
                
                if (byte & 0x04)
                {
                    registerValue |= (1 << D6Pin);
                }
                
                if (byte & 0x02)
                {
                    registerValue |= (1 << D5Pin);
                }
                
                if (byte & 0x01)
                {
                    registerValue |= (1 << D4Pin);
                }
            }

            static void SetBacklight(bool const isOn)
            {
                if (isOn)
                {
                    registerValue |= _BV(BacklightPin);
                }
                else
                {
                    registerValue &= ~_BV(BacklightPin);
                }

                RegisterT::Out(registerValue);
            }

            static uint8_t In()
            {
                return 0;
            }
    };

    template <class RegisterT, uint8_t RsPin, uint8_t RwPin, uint8_t EnPin, uint8_t D4Pin, uint8_t D5Pin, uint8_t D6Pin, uint8_t D7Pin, uint8_t BacklightPin>
    uint8_t Hd44780TransportRegister<RegisterT, RsPin, RwPin, EnPin, D4Pin, D5Pin, D6Pin, D7Pin, BacklightPin>::registerValue = 0;
}
