#include "display_utils.hpp"
#include <avr/pgmspace.h>
#include "board.hpp"
#include "math_utils.hpp"

using namespace Voltmeter;

static uint8_t const CharWaveFormAcData[8] PROGMEM =
{
    0b00000,
    0b00000,
    0b01000,
    0b10101,
    0b00010,
    0b00000,
    0b00000,
    0b00000,
};

static uint8_t const CharWaveFormDcData[8] PROGMEM =
{
    0b00000,
    0b00000,
    0b00000,
    0b11111,
    0b00000,
    0b10101,
    0b00000,
    0b00000,
};

static uint8_t const CharWaveFormAcDcData[8] PROGMEM =
{
    0b00000,
    0b01000,
    0b10101,
    0b00010,
    0b00000,
    0b10101,
    0b00000,
    0b00000,
};

static uint8_t const CharDottedLineData[8] PROGMEM =
{
    0b00000,
    0b00000,
    0b00000,
    0b00000,
    0b00000,
    0b00000,
    0b10101,
    0b00000,
};

static uint8_t const CharSpeakerData[8] PROGMEM =
{
    0b00011,
    0b00101,
    0b11001,
    0b10001,
    0b11001,
    0b00101,
    0b00011,
    0b00000,
};

static uint8_t const CharCheckersPatternData[8] PROGMEM =
{
    0b10101,
    0b01010,
    0b10101,
    0b01010,
    0b10101,
    0b01010,
    0b10101,
    0b01010
};

static uint8_t const CharVolume0Data[8] PROGMEM =
{
    0b00000,
    0b00000,
    0b01010,
    0b00100,
    0b01010,
    0b00000,
    0b00000,
    0b00000,
};

static uint8_t const CharVolume1Data[8] PROGMEM =
{
    0b00000,
    0b00000,
    0b10000,
    0b10000,
    0b10000,
    0b00000,
    0b00000,
    0b00000
};

static uint8_t const CharVolume2Data[8] PROGMEM =
{
    0b00000,
    0b00100,
    0b10100,
    0b10100,
    0b10100,
    0b00100,
    0b00000,
    0b00000
};

static uint8_t const CharVolume3Data[8] PROGMEM =
{
    0b00001,
    0b00101,
    0b10101,
    0b10101,
    0b10101,
    0b00101,
    0b00001,
    0b00000
};

static uint8_t const CharAutoRangeData[8] PROGMEM =
{
    0b00000,
    0b11111,
    0b10001,
    0b10101,
    0b10001,
    0b10101,
    0b11111,
    0b00000
};

static uint8_t const CharBacklightLockData[8] PROGMEM =
{
    0b00000,
    0b11111,
    0b10001,
    0b10101,
    0b10001,
    0b10111,
    0b11111,
    0b00000
};

static uint8_t const CharWarningData[8] PROGMEM =
{
    0b00000,
    0b11111,
    0b11011,
    0b11011,
    0b11111,
    0b11011,
    0b11111,
    0b00000
};


void DisplayUtils::LoadChars()
{
    Board::Display::Driver::LoadCharProgmem(CharBacklightLock, CharBacklightLockData);
    Board::Display::Driver::LoadCharProgmem(CharWaveForm, CharWaveFormDcData);
    Board::Display::Driver::LoadCharProgmem(CharWarning, CharWarningData);
    Board::Display::Driver::LoadCharProgmem(CharDottedLine, CharDottedLineData);
    Board::Display::Driver::LoadCharProgmem(CharSpeaker, CharSpeakerData);
    Board::Display::Driver::LoadCharProgmem(CharVolume, CharVolume3Data);
    Board::Display::Driver::LoadCharProgmem(CharAutoRange, CharAutoRangeData);
}


void DisplayUtils::LoadVolumeChar(uint8_t const level)
{
    uint8_t const * data = CharVolume0Data;

    switch (level)
    {
        case 0:
        default:
            data = CharVolume0Data;
            break;

        case 1:
            data = CharVolume1Data;
            break;

        case 2:
            data = CharVolume2Data;
            break;

        case 3:
            data = CharVolume3Data;
            break;
    }

    Board::Display::Driver::LoadCharProgmem(CharVolume, data);
}


void DisplayUtils::LoadWaveFormChar(WaveForm const waveForm)
{
    uint8_t const * data = CharWaveFormDcData;

    switch (waveForm)
    {
        case WaveForm::Ac:
            data = CharWaveFormAcData;
            break;

        case WaveForm::Dc:
        default:
            data = CharWaveFormDcData;
            break;

        case WaveForm::AcDc:
            data = CharWaveFormAcDcData;
            break;
    }

    Board::Display::Driver::LoadCharProgmem(CharWaveForm, data);
}


void DisplayUtils::DisplayVoltage(Voltage const & voltage)
{
    int16_t value = voltage.value;
    uint8_t digits = MathUtils::Log10(voltage.denominator);

    if (value < 0)
    {
        value = -value;
        Board::Display::PutChar('-');
    }
    else
    {
        Board::Display::PutChar(' ');
    }

    if (voltage.IsOverRange())
    {
        for (int i = 0; i < digits; ++i)
        {
            Voltmeter::Board::Display::PutChar(CharDottedLine);

            if ((voltage.decimalPoint == i) && (i != digits - 1))
            {
                Voltmeter::Board::Display::PutChar('.');
            }
        }
    }
    else
    {
        long position = voltage.denominator / 10;
        bool leadingZeros = true;

        for (int i = 0; i < digits; ++i)
        {
            int digit = value / position;
            value = value % position;
            position = position / 10;

            leadingZeros = leadingZeros && (digit == 0) && (voltage.decimalPoint != i);

            if (leadingZeros)
            {
                Voltmeter::Board::Display::PutChar(' ');
            }
            else
            {
                Voltmeter::Board::Display::PutInt(digit);
            }

            if ((voltage.decimalPoint == i) && (i != digits - 1))
            {
                Voltmeter::Board::Display::PutChar('.');
            }
        }
    }

    Voltmeter::Board::Display::PutChar(' ');

    if (voltage.unit == Voltage::Unit::mV)
    {
        Voltmeter::Board::Display::PutChar('m');
    }

    Voltmeter::Board::Display::PutChar('V');
}